import define1 from "./a33468b95d0b15b0@806.js";

export default function define(runtime, observer) {
  const main = runtime.module();
  const fileAttachments = new Map([["Papers, Taxonomy, and Elicitation Paths - Taxonomy@5.csv",new URL("./files/9c6b6d74cb5d249958fe6f0c22ac537ae22924dc59e1e8c481bdf6f8f380a0fc0e23269590cf4133b009577198d1e5626106edf9afeaae37e22c296f195605a8",import.meta.url)],["Papers, Taxonomy, and Elicitation Paths - Elicitation Paths@5.csv",new URL("./files/e46e78a671fa7b635f4a2cb605db6591e3166e361dbe4fd1b48f885b400f8f33705c26bc60be7c7c83a8688f4a9dc5d67f57a5f0266c1f277651afe2ea8d3ca1",import.meta.url)]]);
  main.builtin("FileAttachment", runtime.fileAttachments(name => fileAttachments.get(name)));
  main.variable(observer()).define(["md"], function(md){return(
md`# A Survey of Domain Knowledge Elicitation in Applied Machine Learning

[Link to Observable notebook](https://observablehq.com/d/af1f5d9971c19aa9).`
)});
  main.variable(observer()).define(["md"], function(md){return(
md`## Counts by purpose`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({
  // the elicitation goals are shown in the columns,
  // so we don't need it in the color scale
  color: color.copy()
      .domain(color.domain().slice(1))
      .range(color.range().slice(1))
})
)});
  main.variable(observer()).define(["width","d3","goals","maxCount","lowLevelNoGoals","countsByGoal","abbreviations","taxonomy","color","lowToTop"], function(width,d3,goals,maxCount,lowLevelNoGoals,countsByGoal,abbreviations,taxonomy,color,lowToTop)
{
  // set up

  const margin = {top: 30, right: 5, bottom: 10, left: 200};
  const visWidth = width - margin.left - margin.right;
  const visHeight = 600 - margin.top - margin.bottom;
  
  const svg = d3.create('svg')
      .attr('width', visWidth + margin.left + margin.right)
      .attr('height', visHeight + margin.top + margin.bottom);
  
  const g = svg.append('g')
      .attr('transform', `translate(${margin.left}, ${margin.top})`);
  
  // scales
  
  // positions the columns of bar charts
  const group = d3.scaleBand()
      .domain(goals)
      .range([0, visWidth])
      .paddingInner(0.05);
  
  // width of bars
  const x = d3.scaleLinear()
      .domain([0, maxCount]).nice()
      .range([0, group.bandwidth()]);

  // vertical position of bars
  const y = d3.scaleBand()
      .domain(lowLevelNoGoals)
      .range([0, visHeight])
      .padding(0.2);
  
  // group for each column (elicitation goal)
  
  const groups = g.selectAll('.group')
    .data(countsByGoal)
    .join('g')
      .attr('class', 'group')
      .attr('transform', ([goal, lowToCount]) => `translate(${group(goal)})`);

  // axes

  // y-axis
  
  // add one y-axis for the codes
  
  const yAxis = d3.axisLeft(y)
      .tickSize(0)
      .tickFormat(d => abbreviations.get(d) ?? d);

  const yAxisGroup = g.append("g")
      .call(yAxis)
      .call(g => g.selectAll('.domain').remove());
  
  // add a y-axis to each column, but only the domain line
  
  groups.append('g')
      .call(yAxis)
      .call(g => g.selectAll('.tick').remove())
      .call(g => g.selectAll('.domain').style('stroke', '#aaa'));

  // mid level labels on y-axis
  
  
  // get the y coordinates where the lines for each level group should start and end
  // slice skips elicitation goals
  const mids = taxonomy.slice(1).map(({mid, lows}) => ({
    mid,
    first: y(lows[0]) + (y.bandwidth() / 2),
    last: y(lows[lows.length - 1]) + (y.bandwidth() / 2),
  }));
  
  // labels
  yAxisGroup.append("g")
    .selectAll('text')
    .data(mids)
    .join('text')
      .attr('x', -100)
      .attr('y', d => (d.first + d.last) / 2)
      .attr('dominant-baseline', 'middle')
      .attr('text-anchor', 'end')
      .attr('font-family', 'sans-serif')
      .attr('font-size', '10')
      .attr('fill', 'black')
      .text(d => d.mid);
  
  // domain line
  yAxisGroup.append("g")
    .selectAll('line')
    .data(mids)
    .join('line')
      .attr('stroke', '#aaa')
      .attr('stroke-linecap', 'round')
      .attr('x1', -90)
      .attr('x2', -90)
      .attr('y1', d => d.first)
      .attr('y2', d => d.last);
  
  // tick marks at the end of the lines
  yAxisGroup.append("g")
    .selectAll('line')
    .data(mids.map(d => [d.first, d.last]).flat())
    .join('line')
      .attr('stroke', '#aaa')
      .attr('stroke-linecap', 'round')
      .attr('x1', -90)
      .attr('x2', -85)
      .attr('y1', d => d)
      .attr('y2', d => d);
  
  // x-axis and goal label

  const xAxis = d3.axisTop(x).ticks(8);
  
  groups.append('g')
      .call(xAxis)
      .call(g => g.selectAll('.domain').style('stroke', '#aaa'))
      .call(g => g.selectAll('.tick>line').style('stroke', '#aaa'))
    .append('text')
      .attr('dominant-baseline', 'middle')
      .attr('text-anchor', 'middle')
      .attr('font-family', 'sans-serif')
      .attr('fill', 'black')
      .attr('x', group.bandwidth() / 2)
      .attr('y', -25)
      .text(([goal, lowToCount]) => goal);

  // bars
  
  groups.selectAll('rect')
    .data(([goal, lowToCount]) => lowToCount)
    .join('rect')
      .attr('fill', ([low, count]) => color(lowToTop.get(low)))
      .attr('y', ([low, count]) => y(low))
      .attr('height', y.bandwidth())
      .attr('x', 0)
      .attr('width', ([low, count]) => x(count))
    .append("title")
      .text(([low, count]) => `${low}\n${count}`)
  
  return svg.node();
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`## All paths`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({ color })
)});
  main.variable(observer()).define(["drawSankey","getSankeyData","rows"], function(drawSankey,getSankeyData,rows){return(
drawSankey(getSankeyData(rows))
)});
  main.variable(observer()).define(["md"], function(md){return(
md`## Problem specification`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({ color })
)});
  main.variable(observer()).define(["rows","drawSankey","getSankeyData"], function(rows,drawSankey,getSankeyData)
{
  const filteredRows = rows.filter(d => d['Problem specification'] === 'x');
  filteredRows.columns = rows.columns;
  return drawSankey(getSankeyData(filteredRows));
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`## Feature engineering`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({ color })
)});
  main.variable(observer()).define(["rows","drawSankey","getSankeyData"], function(rows,drawSankey,getSankeyData)
{
  const filteredRows = rows.filter(d => d['Feature engineering'] === 'x');
  filteredRows.columns = rows.columns;
  const offsets = new Map([
    ['Validated', -15],
    ['Unvalidated', -15]
  ]);
  return drawSankey(getSankeyData(filteredRows), offsets);
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`## Model development`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({ color })
)});
  main.variable(observer()).define(["rows","drawSankey","getSankeyData"], function(rows,drawSankey,getSankeyData)
{
  const filteredRows = rows.filter(d => d['Model development'] === 'x');
  filteredRows.columns = rows.columns;

  const offsets = new Map([
    ['Constrained', 15],
    ['Unvalidated', -15]
  ]);

  return drawSankey(getSankeyData(filteredRows), offsets);
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`## Model evaluation`
)});
  main.variable(observer()).define(["swatches","color"], function(swatches,color){return(
swatches({ color })
)});
  main.variable(observer()).define(["rows","drawSankey","getSankeyData"], function(rows,drawSankey,getSankeyData)
{
  const filteredRows = rows.filter(d => d['Model evaluation'] === 'x');
  filteredRows.columns = rows.columns;
  
  const offsets = new Map([
    ['Unvalidated', -10]
  ]);
  
  return drawSankey(getSankeyData(filteredRows), offsets);
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`## Sankey Setup`
)});
  main.variable(observer("abbreviations")).define("abbreviations", function(){return(
new Map(Object.entries({
  // elicitation goals
  'Problem specification': 'Problem spec.',
  'Feature engineering': 'Ft. eng.',
  'Model development': 'Model dev.',
  'Model evaluation': 'Model eval.',
  // elicitation targets
  'Background knowledge and processes': 'Background',
  'Labeling functions': 'Labeling func.',
  'Label explanations': 'Label explan.',
  'Probability distribution': 'Prob. dist.',
  'Feature explanations': 'Ft. explanations',
  'Feature relationships': 'Ft. relationships',
  'Feature relevance': 'Ft. relevance',
  'Feature transformations': 'Ft. transform.',
  'Feature weights': 'Ft. weights',
  'Feature direction': 'Ft. direction',
  'Equivalent sample size': 'Eq. sample size',
  'Model constraints': 'Constraints',
  // elicitation process
  'Meetings or interviews': 'Mtgs. / interviews',
  'Well-defined prompts': 'Well-defined',
  'Not well-defined prompts': 'Not well-defined',
  'Response format': 'Resp. format',
  'Response validation': 'Resp. validation',
  'Unreliable responses': 'Unreliable resp.',
  'Not accounted for': 'Not acct. for',
  'Number of experts': 'Num. experts',
}))
)});
  main.variable(observer("drawSankey")).define("drawSankey", ["d3","width","height","color","abbreviations","midLevel","margins"], function(d3,width,height,color,abbreviations,midLevel,margins){return(
function drawSankey(data, labelOffsets=null) {
  // create outer svg element and set its dimensions
  const svg = d3.create("svg")
      .attr("width", width)
      .attr("height", height);

  // nodes
  
  // nodes can still have a value of 0 despite having filtered
  // the links with value of 0 in getLinks
  const nodesWithValues = data.nodes.filter(d => d.value !== 0);
  
  svg.append("g")
    .selectAll("rect")
    .data(nodesWithValues)
    .join("rect")
      .attr("x", d => d.x0)
      .attr("y", d => d.y0)
      .attr("height", d => d.y1 - d.y0)
      .attr("width", d => d.x1 - d.x0)
      .attr("fill", d => color(d.top))
    .append("title")
      .text(d => `${d.low}\n${d.value}`);

  // links
  
  const link = svg.append("g")
      .attr("fill", "none")
      .attr("stroke-opacity", 0.5)
    .selectAll("g")
    .data(data.links)
    .join("g")
      .style("mix-blend-mode", "multiply");

  link.append("path")
      .attr("d", d3.sankeyLinkHorizontal())
      .attr("stroke", "#aaa")
      // min stroke-width is 1
      .attr("stroke-width", d => Math.max(1, d.width));

  link.append("title")
      .text(d => `${d.source.low} → ${d.target.low}\n${d.value}`);

  // node labels
  
  svg.append("g")
      .attr("font-family", "sans-serif")
      .attr("font-size", 10)
    .selectAll("text")
    .data(nodesWithValues)
    .join("text")
      // labels for nodes on the left half are to the right of the node
      // labels for nodes on the right half are to the left of the node
      .attr("x", d => d.x0 < (width / 2) ? d.x1 + 3 : d.x0 - 3)
      // vertically center the label with the node, but apply
      // offset as needed to avoid overlap
      .attr("y", d => {
        const offset = labelOffsets?.get(d.low) ?? 0;
        return offset + ((d.y1 + d.y0) / 2);
      })
      .attr("dy", "0.35em")
      .attr("text-anchor", d => d.x0 < (width / 2) ? "start" : "end")
      .text(d => abbreviations.get(d.low) ?? d.low);
  
  // column labels

  // get a map from the column name to the
  // x coordinate of the center of the column
  const sankeyColumnToCx = d3.rollup(
    nodesWithValues,
    /* each node in the group should have the same x0 and x1 attributes,
       so we can just use the first node to get the center of the column */
    g => (g[0].x0 + g[0].x1) / 2,
    // group nodes by mid level category (column in sankey)
    d => d.mid
  )
  
  svg.append("g")
      .attr("font-family", "sans-serif")
      .attr("font-size", 10)
    .selectAll("text")
    .data(midLevel)
    .join("text")
      .attr("x", d => sankeyColumnToCx.get(d))
      .attr("y", margins.top / 2)
      .attr("text-anchor", "middle")
      .attr("dominant-baseline", "middle")
      .text(d => abbreviations.get(d) ?? d);

  // return the svg element to render it
  return svg.node();
}
)});
  main.variable(observer("color")).define("color", ["d3","topLevel"], function(d3,topLevel){return(
d3.scaleOrdinal()
    .domain(topLevel)
    .range(d3.schemeCategory10)
)});
  main.variable(observer("width")).define("width", function(){return(
954
)});
  main.variable(observer("height")).define("height", function(){return(
450
)});
  main.variable(observer("margins")).define("margins", function(){return(
{ top: 25, bottom: 10, left: 35, right: 35 }
)});
  main.variable(observer()).define(["md"], function(md){return(
md`## Data processing

Load the attached file and split it into lines`
)});
  main.variable(observer("spreadsheet")).define("spreadsheet", ["FileAttachment"], async function(FileAttachment){return(
(await FileAttachment("Papers, Taxonomy, and Elicitation Paths - Elicitation Paths@5.csv").text()).split("\n")
)});
  main.variable(observer()).define(["md"], function(md){return(
md`Turn the lines for the eliciation paths (rows) into objects.`
)});
  main.variable(observer("rows")).define("rows", ["spreadsheet","d3"], function(spreadsheet,d3)
{
  // remove first two header lines and last two summary count lines
  // this leaves only the rows for the papers
  const text = spreadsheet.slice(2, -2).join("\n");
  
  return d3.csvParse(text, d => {
    // remove empty key
    delete d[""];
    return d;
  })
}
);
  main.variable(observer()).define(["md"], function(md){return(
md`The taxonomy has three levels in the hierarchy. Here we get the categories in the top two levels.`
)});
  main.variable(observer("topLevel")).define("topLevel", ["spreadsheet"], function(spreadsheet){return(
spreadsheet[0].split(",").filter(d => d != "").slice(0, 4)
)});
  main.variable(observer("midLevel")).define("midLevel", ["spreadsheet"], function(spreadsheet){return(
spreadsheet[1].split(",").filter(d => d != "").slice(0, 11)
)});
  main.variable(observer()).define(["md","taxonomy"], function(md,taxonomy){return(
md`
\`nodes\` contains one object each low level category. Each object contains the low level category and the mid and top level categories that it is under.

\`taxonomy\` contains one object for each mid level category. Each object contains the mid level category, the top level category that it is under, and low level categories that are beneath it.

In the Sankey diagrams, each node represents one low level category. The nodes are horizontally aligned based on the mid level category that they are under. This means our Sankey diagram will have ${taxonomy.length} columns, one for each mid level category.`
)});
  main.variable(observer("nodes")).define("nodes", ["FileAttachment"], function(FileAttachment){return(
FileAttachment("Papers, Taxonomy, and Elicitation Paths - Taxonomy@5.csv").csv()
)});
  main.variable(observer("taxonomy")).define("taxonomy", ["d3","nodes"], function(d3,nodes){return(
d3.rollups(
  nodes,
  // each group of nodes has the same top and mid level category
  group => ({
      top: group[0].top,
      mid: group[0].mid,
      lows: group.map(d => d.low)
  }),
  // group the nodes by the mid level category
  d => d.mid
).map(([key, value]) => value)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`\`getLinks(rows)\` is a function that returns the data for the edges in a Sankey diagram. It has one parameter, \`rows\`, which is an arrray of rows in the spreadsheet. Each row represents one elicitation path. It uses [\`d3.pairs\`](@d3/d3-pairs) to get all pairs of adjacent columns (mid level categories) in the diagram.`
)});
  main.variable(observer()).define(["d3","taxonomy"], function(d3,taxonomy){return(
d3.pairs(taxonomy)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`For each pair of columns, it gets all possible pairs of nodes between the two columns. This array is flattened to get one array of objects, where each object represents a possible edge in the Sankey diagram.`
)});
  main.variable(observer("exampleNodes")).define("exampleNodes", ["d3","taxonomy"], function(d3,taxonomy){return(
d3.pairs(taxonomy, (a, b) =>
    d3.cross(a.lows, b.lows, (source, target) => ({ source, target }))
).flat()
)});
  main.variable(observer()).define(["md"], function(md){return(
md`Next, for each posssible edge, this function calculates the number of rows that are represented by the edge. For example, if the edge is from "Problem specification" to "Labels", then we want to count the number of rows that have x's for both "Problem specification" and "Labels". We can then filter out any edges that have a count of 0, which means they won't be in the Sankey diagram.`
)});
  main.variable(observer("exampleLinks")).define("exampleLinks", ["exampleNodes","rows"], function(exampleNodes,rows){return(
exampleNodes.map(({ source, target }) => {
    // get number of rows where both columns have an x
    const value = rows
      .filter(d => d[source] === "x" && d[target] === "x")
      .length;

    return { source, target, value };
  })
  .filter(d => d.value != 0)
)});
  main.variable(observer("getLinks")).define("getLinks", ["d3","taxonomy"], function(d3,taxonomy){return(
function getLinks(rows) {
  // get list of pairwise columns
  return d3.pairs(taxonomy, (a, b) =>
    // cross-product of categories in those columns
    d3.cross(a.lows, b.lows, (source, target) => ({ source, target }))
  )
  .flat()
  .map(({ source, target }) => {
    // get number of rows where both columns have an x
    const value = rows
      .filter(d => d[source] === "x" && d[target] === "x")
      .length;

    return { source, target, value };
  })
  .filter(d => d.value !== 0);
}
)});
  main.variable(observer()).define(["md"], function(md){return(
md`\`getSankeyData()\` creates a Sankey generator and passes the nodes and links to the generator.`
)});
  main.variable(observer("getSankeyData")).define("getSankeyData", ["d3","margins","width","height","nodes","getLinks"], function(d3,margins,width,height,nodes,getLinks){return(
function getSankeyData(rows) {
  const sankey = d3.sankey()
      .nodeId(d => d.low)
      .nodeWidth(5)
      .nodePadding(15)
      // [[x0, y0], [x1, y1]]
      .extent([[margins.left, margins.top],
               [width - margins.right, height - margins.bottom]]);
  
  return sankey({
    // create copies of the nodes and links so that the original values aren't mutated
    nodes: nodes.map(d => Object.assign({}, d)),
    links: getLinks(rows).map(d => Object.assign({}, d))
  });
}
)});
  main.variable(observer()).define(["md","goals"], function(md,goals){return(
md`### Bar chart data processing

For the bar chart, we need an array of all of the low level categories in the taxonomy, excluding the elicitation goals. These categories are the rows in the bar chart. The ${goals.length} goals are the columns.`
)});
  main.variable(observer("lowLevelNoGoals")).define("lowLevelNoGoals", ["nodes"], function(nodes){return(
nodes.map(d => d.low).slice(4)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`Map from the categories in the low level of the taxonomy to the top level category that they are under.`
)});
  main.variable(observer("lowToTop")).define("lowToTop", ["taxonomy"], function(taxonomy){return(
new Map(
  taxonomy.map(
    ({top, lows}) => lows.map(low => [low, top])
  ).flat()
)
)});
  main.variable(observer("goals")).define("goals", ["rows"], function(rows){return(
rows.columns.slice(2, 6)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`Get the number of times each low level category appears with each elicitation goal.`
)});
  main.variable(observer("countsByGoal")).define("countsByGoal", ["d3","rows","lowLevelNoGoals"], function(d3,rows,lowLevelNoGoals){return(
d3.rollup(
  rows,
  // g is an array containing all of the rows that have the same elicitation goal
  g => g.reduce(
      /*  the first argument to reduce is a reducer function.
          this function has two arguments:
          - accumulator
          - current value
      */
      (acc, cur) => {
        // loop over the bottom level categories
        // (columns in the spreadsheet)
        lowLevelNoGoals.forEach(col => {
          // if the current row has an "x" for this column,
          // then increment the count for this column in the
          // acumulator
          if (cur[col] === "x") {
            const count = acc.get(col) ?? 0;
            acc.set(col, count + 1);
          }
        });
        // the returned accumulator will be passed to the
        // next call of the reducer
        return acc;
      },
      // the second argument to reduce is an initial value
      // for the accumulator
      new Map()
    ),
  // group the rows according to which elicitation goal is marked
  d => {
    if (d["Problem specification"] === "x") {
      return "Problem specification";
    } else if (d["Feature engineering"] === "x") {
      return "Feature engineering";
    } else if (d["Model development"] === "x") {
      return "Model development";
    } else if (d["Model evaluation"] === "x") {
      return "Model evaluation";
    }
  }
)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`Get the highest count in \`countsByPurpose\`.`
)});
  main.variable(observer("maxCount")).define("maxCount", ["d3","countsByGoal"], function(d3,countsByGoal){return(
d3.max(
  countsByGoal,
  ([purpose, categoryToCount]) => d3.max(categoryToCount, ([category, count]) => count)
)
)});
  main.variable(observer()).define(["md"], function(md){return(
md`----
## Appendix`
)});
  const child1 = runtime.module(define1);
  main.import("swatches", child1);
  main.variable(observer("d3")).define("d3", ["require"], function(require){return(
require("d3@7", "d3-sankey@0.12")
)});
  return main;
}
